#include "hmbdc/Copyright.hpp"
#pragma once

#include <string>
#include <memory>
#include <list>
#include <unordered_set>
#include <boost/property_tree/ptree.hpp>
#pragma GCC diagnostic ignored "-Wreturn-type"
#include <boost/property_tree/json_parser.hpp>

namespace hmbdc { namespace app { 

using namespace std;
using namespace boost::property_tree;
/**
 * @brief class to hold an hmbdc configuration
 * @details it is based on a two level json format.
 * top level for the fallback values and lower level for section specific values
 * shown below:
 * 
 *      {
 *          "parameter_1": "top level, value used as a fallback",
 *          "parameter_2": "fallback is used when not configured in a section",
 *          "section_A": {
 *               "parameter_2": "lower level, a specific value effective in section_A",
 *               "parameter_1": "a specific value effective in section_A"
 *          },
 *          "another_section": {
 *               "parameter_2": "a specific value effective in another_section"
 *          }
 *      }
 *      
 */
struct Config 
: ptree {
    using ptree::ptree;
    /**
     * @brief empty config
     */
    Config(){}

    /**
     * @brief construct using stream, optionally specifying the section name
     * @details if the section is nullptr, just use the fallback values. if
     * the section name cannot be found, throw an exception
     * 
     * @param is stream as input providing a json stream
     * @param section pointing to the effective section in the json above
     */
    explicit 
    Config(istream&& is, char const* section = nullptr) {
        if (section) {
            read_json(is, fallback_);
            (ptree&)*this = fallback_.get_child(section);
        } else {
            read_json(is, (ptree&)*this);
        }
    }
    /**
     * @brief construct using stream, optionally specifying the section name
     * @details if the section is nullptr, just use the fallback values. if
     * the section name cannot be found, throw an exception
     * 
     * @param is stream as input providing a json stream
     * @param section pointing to the effective section in the json above
     */
    explicit 
    Config(istream& is, char const* section = nullptr) {
        if (section) {
            read_json(is, fallback_);
            (ptree&)*this = fallback_.get_child(section);
        } else {
            read_json(is, (ptree&)*this);
        }
    }

    /**
     * @brief construct using a string, optionally specifying the section name
     * @details if the section is nullptr, just use the fallback values. if
     * the section name cannot be found, throw an exception
     * 
     * @param json string as input providing a json text
     * @param section pointing to the effective section in the json above
     */
    explicit 
    Config(char const* json, char const* section = nullptr)
    : Config(istringstream(json), section)
    {}

    /**
     * @brief construct using another ptree as fallbacks, optionally specifying the section name
     * @details if the section is nullptr, just use the fallback values. if
     * the section name cannot be found, throw an exception
     * 
     * @param t ptree as input providing fallbacks (and sections)
     * @param section pointing to the effective section in the ptree above
     */
    Config(ptree const& t, char const* section = nullptr) {
        if (section) {
            fallback_ = t;
            (ptree&)*this = fallback_.get_child(section);
        } else {
            (ptree&)*this = t;
        }
    }

    /**
     * @brief construct using a fallback ptree, and specify the section ptree
     * @details if the section is nullptr, just use the fallback values. if
     * the section name cannot be found, throw an exception
     * 
     * @param t ptree as input providing fallbacks
     * @param section ptree providing section specific values
     */
    explicit
    Config(ptree const& t, ptree const& section) {
        fallback_ = t;
        (ptree&)*this = section;
    }

    /**
     * @brief internal use
     * @details this is to set the default user config values (not the fallback values)
     * 
     * @param c a config holding configuration values
     */
    void setDefaultUserConfig(Config const& c) {
        defaultUserConfig_.reset(new Config(c));
    }

    /**
     * @brief get a value from the config
     * @details check the section for it, if not found, try use fallback provided
     * if still missing, search using the default user config values set by
     * setDefaultUserConfig. Throw exception ptree_bad_path if all fail
     * 
     * @param param config parameter name
     * @tparam T type of the value
     * @return result
     */
    template <typename T>
    T getExt(const path_type& param) const {
        auto res = get_optional<T>(param);
        if (!res) {
            res = fallback_.get_optional<T>(param);
            if (!res) {
                if (defaultUserConfig_) {
                    return defaultUserConfig_->getExt<T>(param);
                } else {
                    throw ptree_bad_path("invalid param and no default user Config set", param); 
                }
            }
        }
        return *res; 
    }

    /**
     * @brief get a number value in hex format
     * @details check the section for it, if not found, try use fallback provided
     * if still missing, search using the default user config values set by
     * setDefaultUserConfig. Throw exception ptree_bad_path if all fail
     * 
     * @param param config parameter name
     * @tparam T numeric type of the value: int , uint64_t ...
     * @return result
     */
    template <typename T>
    T getHex(ptree::path_type const& param) {
        istringstream iss(getExt<string>(param));
        T res;
        iss >> hex >> res;
        return res;
    }
 
    /**
     * @brief fill in a variable with a configured value retrieved using getExt
     * @details example cfg(abc, "abc")(def, "def");
     * 
     * @param to destination
     * @param param config parameter
     * 
     * @return the Config object itself
     */
    template <typename T>
    Config const& operator()(T& to, const path_type& param) const {
        to = getExt<T>(param);
        return *this;
    }

    /**
     * @brief get contents of all  the effective configure in the form of list of string pairs
     * @details only effective ones are shown
     * 
     * @param skipThese skip those config params
     * @return list of string pairs in the original order of ptree nodes
     */
    list<pair<string, string>> content(
        unordered_set<string> const& skipThese = unordered_set<string>()) const {
         list<pair<string, string>> res;
         unordered_set<string> history(skipThese);
         for (auto& p : *this) {
            if (history.find(p.first) == history.end()) {
                history.insert(p.first);
                if (p.second.empty()) {
                    res.push_back(make_pair(p.first, p.second.get_value<string>()));
                }
            }
         }
         for (auto& p : fallback_) {
            if (history.find(p.first) == history.end()) {
                history.insert(p.first);
                if (p.second.empty()) {
                    res.push_back(make_pair(p.first, p.second.get_value<string>()));
                }
            }
         }
         if (defaultUserConfig_) {
            auto more = defaultUserConfig_->content(history);
            res.insert(res.end(), more.begin(), more.end());
         }
         return res;
    }

private:
    ptree fallback_;
    std::shared_ptr<Config> defaultUserConfig_;
};

}}

