#include "hmbdc/Copyright.hpp"
#pragma once

#include "hmbdc/time/Time.hpp"
#include <stdint.h>
#include <utility>
#include <ostream>
#include <type_traits>
#include <memory.h>

namespace hmbdc { namespace app {

/**
 * @brief each message type has 16 bit tag
 * @details 0-1000 is reserved for internal use, user message starts from 1001
 * 
 * see @example hmbdc.cpp for defining messages
 * @tparam tag 16 bit unsigned tag
 */
template <uint16_t tag>
struct hasTag {
    enum {typeTag = tag};
};

struct MessageHead {
	MessageHead(uint16_t typeTag)
	: reserved2(0)
    , reserved(0)
	, typeTag(typeTag)
	{}
	template <typename Message> Message& get();
	template <typename Message> Message const& get() const;
	template <typename Message> void setPayload(Message const&);
	template <typename Message, typename ...Args> void setPayloadInPlace(Args&& ... args);
    uint32_t reserved2;
    uint16_t reserved;
    uint16_t typeTag;

    friend 
    std::ostream& operator << (std::ostream& os, MessageHead const & h) {
    	return os << h.reserved << ' ' << h.typeTag;
    }
} __attribute__((packed));

template <typename Message>
struct MessageWrap : MessageHead {
    template <typename ...Args>
	MessageWrap(Args&& ... args)
	: MessageHead(Message::typeTag)
	, payload(std::forward<Args>(args)...) {
		static_assert(std::is_class<Message>::value, "wrong type to wrap");
	}
    Message payload;

    friend 
    std::ostream& operator << (std::ostream& os, MessageWrap<Message> const & w) {
    	return os << static_cast<MessageHead const&>(w) << ' ' << w.payload;
    }
};

template <typename Message>
Message& 
MessageHead::get() {
	return static_cast<MessageWrap<Message>*>(this)->payload;
}

template <typename Message>
Message const& 
MessageHead::get() const{
	return static_cast<MessageWrap<Message> const*>(this)->payload;
}

template <typename Message> 
void 
MessageHead::
setPayload(Message const& m) {
	new (&get<Message>()) Message(m);
	typeTag = Message::typeTag;
}

template <typename Message, typename ...Args> 
void
MessageHead::
setPayloadInPlace(Args&& ... args) {
	new (&get<Message>()) Message(std::forward<Args>(args)...);
	typeTag = Message::typeTag;
}

struct Flush
: hasTag<0> {
};

template <size_t MaxStreamableSize>
struct LoggingT
: hasTag<3> {
    char payload[MaxStreamableSize];
} __attribute__((__may_alias__, packed));

/**
 * @brief A special type of message
 * @details When a Client declare this type as interested, all
 * messages will be delievered in byte form to this client
 * additonally - see Client.hpp and ConsoleClient.hpp documentation
 */
struct JustBytes
: hasTag<4> {
	JustBytes() = delete;
};

struct Trace {
    Trace() 
    : tsIndex(1) {
        ts[0] = (time::SysTime::now());
    }
    uint8_t tsIndex;
    time::SysTime ts[6];
    friend 
    std::ostream& operator << (std::ostream& os, Trace const & t) {
        for (auto i = 0u; i < t.tsIndex; ++i) os << t.ts[i] << ' ';
        return os;
    }
};

template<>
struct MessageWrap<JustBytes> :  MessageHead {
	MessageWrap(uint16_t tag, void const* bytes, size_t len)
	: MessageHead(tag) {
		memcpy(&payload, bytes, len);
	}
    uint8_t payload;

    friend 
    std::ostream& operator << (std::ostream& os, MessageWrap<JustBytes> const & w) {
    	return os << static_cast<MessageHead const&>(w) << " *";
    }
};
static_assert(sizeof(MessageWrap<JustBytes>) == sizeof(MessageHead) + 1, "wrong packing");
static_assert(sizeof(MessageWrap<Flush>) == sizeof(MessageHead) + sizeof(Flush), "wrong packing");
}} //hmbdc::app
