#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/Config.hpp"
#include "hmbdc/app/mcast/Messages.hpp"
#include "hmbdc/app/mcast/Sender.hpp"
#include "hmbdc/app/mcast/SendTransportEngine.hpp"
#include "hmbdc/app/mcast/RecvTransportEngine.hpp"
#include "hmbdc/app/mcast/DefaultUserConfig.hpp"
#include "hmbdc/comm/Topic.hpp"
#include "hmbdc/pattern/GuardedSingleton.hpp"

#include <boost/regex.hpp>

#include <vector>
#include <mutex>
#include <memory>

//commands used to build and install netmap for intel NIC
//./configure --kernel-dir=/usr/src/kernels/3.10.0-327.el7.x86_64/ --kernel-sources=/home/urname/rpmbuild/SOURCES/linux-3.10.0-327.36.3.el7 --drivers=e1000e
//make
//sudo make install


namespace hmbdc { namespace app { namespace mcast {

/**
 * @brief a singleton that holding mcast resources
 * @details it manages transport engines
 * 
 * see @example perf-mcast.cpp and @example server-cluster.cpp for usage.
 * 
 */
struct NetContext 
: hmbdc::pattern::GuardedSingleton<NetContext> {

    friend class hmbdc::pattern::SingletonGuardian<NetContext>;
    using Sender = hmbdc::app::mcast::Sender;

/**
 * @brief construct a send transport engine (and remember it within the class)
 * @details After this, user is responsible to get it started within a hmbdc Context,
 * if running in Context pool, it needs to pin at a single pool thread, CANNOT span
 * more than one thread
 * otherwise the transport is not functioing/running. Don't create the same thing twice
 * 
 * @param cfgIn config specifing the transport - see example in perf-mcast.cpp and DefaultUserConfig.hpp
 * @param maxMessageSize max messafe size in bytes to be sent
 * trasnport engines, see ctor of SendTransportEngine for details
 * @return a pointer to the Engine - don't delete it
 */
    SendTransportEngine* createSendTransportEngine(Config const& cfgIn
    	, size_t maxMessageSize) {
        Config dft(DefaultUserConfig, "tx");
        Config cfg(cfgIn);
        cfg.setDefaultUserConfig(dft);

        std::lock_guard<std::mutex> tlock(sendTransportsLock_);
        auto res = new SendTransportEngine(cfg, maxMessageSize);
        sendTransports_.emplace_back(res);
        return res;
    }

/**
 * @brief same as above but provide a unified interface - not preferred
 * @return a pointer to the Engine - don't delete it
 */
    SendTransportEngine* createSendTransportEngineTuply(Config const& cfg
        , size_t maxMessageSize
        , tuple<> args) {
        return createSendTransportEngine(cfg, maxMessageSize);
    }

/**
 * @brief construct a send transport and remember it
 * @details After this, user is responsible to get it started within a hmbdc Context,
 * if running in Context pool, it needs to pin at a single pool thread, CANNOT span
 * more than one thread
 * otherwise the transport is not functioing/running. Don't create the same thing twice
 * 
 * @param cfgIn jason specifing the transport - see example, perf-mcast.cpp and DefaultUserConfig.hpp
 * @param buffer buffer that recv messages go, normally the one returned by app::Context::buffer()
 * @param arb optonally an arbitrator to decide which messages to keep and drop
 * if arb is an rvalue, it is passed in value, if an lvalue, passed in as reference; 
 * it supports either udp packet (BEFORE topic filtering) or 
 * hmbdc message (AFTER topic filtering) level arbitration depending on which one of 
 * int operator()(void* bytes, size_t len) and 
 * int operator()(TransportMessageHeader const* header) presents in the arb passed in
 * trasnport engines, see ctor of RecvTransportEngineImpl for details
 * @return a pointer to the Engine
 */
    template <typename Buffer, typename MsgArbitrator = RecvTransport::NoOpArb>
    auto createRecvTransportEngine(Config const& cfgIn
        , Buffer& buffer
        , MsgArbitrator&& arb = RecvTransport::NoOpArb()) {
        Config dft(DefaultUserConfig, "rx");
        Config cfg(cfgIn);
        cfg.setDefaultUserConfig(dft);

        std::lock_guard<std::mutex> tlock(recvTransportsLock_);
        auto res = 
            new RecvTransportEngineImpl<Buffer, MsgArbitrator>(cfg, buffer
                , forward<MsgArbitrator>(arb));
        recvTransports_.emplace_back(res);
        return res;
    }

/**
 * @brief same as above but to provide a unified interface - not preferred
 * @details use forward_as_tuple to make the tuple passed in
 * @return a pointer to the Engine
 */
    template <typename Buffer, typename ArgsTuple>
    auto createRecvTransportEngineTuply(
        Config const& cfg
        , Buffer& buffer
        , ArgsTuple&& args) {
        return createRecvTransportEngine(cfg
            , buffer
            , get<0>(args)
            , get<1>(args)
        );
    }

/**
 * @brief get (or create for the first time) a Sender - whose function is to send messages on 
 * its associated Topic
 * @details this operation typically might be slow, so caching the return value is recommended. 
 * 
 * @param t - the Topic that the Sender is for
 */
    Sender* getSender(Topic const& t) {
        std::lock_guard<std::mutex> lock(sendersLock_);
        auto sender = senders_.find(t);
        if ( sender != senders_.end()) {
            return sender->second.get();
        } else {
            std::lock_guard<std::mutex> slock(sendTransportsLock_);
            for (auto i = 0u;
                i < sendTransports_.size();
                ++i) {
                auto st = sendTransports_[i];
                if (st->match(t)) {
                    auto newSender = new Sender(st, t);
                    senders_[t].reset(newSender);
                    return newSender;
                }
            }
        }

        return nullptr;
    }

/**
 * @brief This process is interested in a Topic
 * @details Normally the receiving transport covering this topic
 * needs to be created - not necessarily running - before calling this 
 * 
 * @param t Topic interested
 */
    void listenTo(Topic const& t) {
        std::lock_guard<std::mutex> tlock(recvTransportsLock_);
        for (auto ptr : recvTransports_) {
            ptr->listenTo(t);
        }
    }
    
/**
 * @brief undo the subscription
 * 
 * @param t Topic
 */
    void stopListenTo(Topic const& t) {
        std::lock_guard<std::mutex> tlock(recvTransportsLock_);
        for (auto ptr : recvTransports_) {
            ptr->stopListenTo(t);
        }
    }

private:
    NetContext() {
    }

    ~NetContext() {
    }


    std::vector<SendTransport::ptr> sendTransports_;
    std::mutex sendTransportsLock_;

    std::vector<RecvTransport::ptr> recvTransports_;
    std::mutex recvTransportsLock_;

    std::map<Topic, Sender::ptr> senders_;
    std::mutex sendersLock_;    
};

}
}}

