#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/Config.hpp"
#include "hmbdc/app/netmap/Messages.hpp"
#include "hmbdc/app/netmap/Sender.hpp"
#include "hmbdc/app/netmap/SendTransportEngine.hpp"
#include "hmbdc/app/netmap/RecvTransportEngine.hpp"
#include "hmbdc/app/netmap/DefaultUserConfig.hpp"

#include "hmbdc/comm/Topic.hpp"
#include "hmbdc/pattern/GuardedSingleton.hpp"

#include <boost/regex.hpp>

#include <vector>
#include <mutex>
#include <memory>

namespace hmbdc { namespace app { namespace netmap {

/**
 * @brief a singleton that holding netmap resources
 * @details it manage transport engines
 * see @example client-server-netmap.cpp @example perf-netmap.cpp for usage example
 */
struct NetContext 
: hmbdc::pattern::GuardedSingleton<NetContext> {

    friend class hmbdc::pattern::SingletonGuardian<NetContext>;
    using Sender = hmbdc::app::netmap::Sender;
/**
 * @brief construct a send transport and remember it
 * @details After this, user is responsible to get it started within a hmbdc Context, 
 * otherwise the transport is not functioing/running. Don't create the same thing twice.
 * it might take several seconds (configured by nmResetWaitSec=2) to construct the engine 
 * to avoid message losses at the beginning - netmap seems to require that
 * 
 * @param cfgIn jason specifing the transport - see example, perf-netmap.cpp and DefaultUserConfig.hpp
 * @param maxMessageSize max messafe size in bytes to be sent
 * @return a pointer to the Engine
 */
    SendTransportEngine* createSendTransportEngine(Config const& cfgIn, size_t maxMessageSize) {
        Config dft(DefaultUserConfig, "tx");
        Config cfg(cfgIn);
        cfg.setDefaultUserConfig(dft);

        std::lock_guard<std::mutex> tlock(sendTransportEnginesLock_);
        auto res = SendTransportEngine::ptr(new SendTransportEngine(cfg, maxMessageSize));
        sendTransportEngines_.push_back(res);
        return res.get();
    }

/**
 * @brief same as above but provide an unified interface - not preferred
 * @return a pointer to the Engine - don't delete it
 */
    SendTransportEngine* createSendTransportEngineTuply(Config const& cfg
        , size_t maxMessageSize
        , tuple<> args) {
        return createSendTransportEngine(cfg, maxMessageSize);
    }

/**
 * @brief construct a send transport and remember it
 * @details After this, user is respoonsible to get it started within a hmbdc Context, 
 * otherwise the transport is not functioing/running. Don't create the same thing twice
 * 
 * @param cfgIn jason specifing the transport - see example, perf-netmap.cpp and DefaultUserConfig.hpp
 * @param buffer buffer that recv messages go in, normally the one returned by 
 * app::Context::buffer()
 * @param arb optonally an arbitrator to decide which messages to keep and drop
 * if arb is an rvalue, it is passed in value, if an lvalue, passed in as reference;
 * it also supports either netmap packet level (BEFORE topic filtering) or 
 * hmbdc message level (AFTER topic filtering)  arbitration depending on which one of 
 * int operator()(void* bytes, size_t len) and 
 * int operator()(TransportMessageHeader const* header) presents in the arb passed in

 * @return a pointer to the Engine
 */
    template <typename Buffer, typename MsgArbitrator = RecvTransport::NoOpArb>
    auto createRecvTransportEngine(Config const& cfgIn
        , Buffer& buffer
        , MsgArbitrator&& arb = RecvTransport::NoOpArb()) {
        Config dft(DefaultUserConfig, "rx");
        Config cfg(cfgIn);
        cfg.setDefaultUserConfig(dft);

        std::lock_guard<std::mutex> tlock(recvTransportEnginesLock_);
        auto res = 
            new RecvTransportEngineImpl<Buffer, MsgArbitrator>(cfg, buffer
                , forward<MsgArbitrator>(arb));
        recvTransportEngines_.emplace_back(res);
        return res;
    }

/**
 * @brief same as above but to provide a unified interface - not preferred 
 * @return a pointer to the Engine
 */
    template <typename Buffer, typename ArgsTuple>
    auto createRecvTransportEngineTuply(
        Config const& cfg
        , Buffer& buffer
        , ArgsTuple&& args) {
        return createRecvTransportEngine(cfg
            , buffer
            , get<0>(args)
        );
    }

/**
 * @brief get (or create for the first time) a Sender - whose function is to send messages on 
 * its associated Topic
 * @details this operation typically might be slow, so caching the return value is recommended. 
 * 
 * @param t - the Topic that the Sender is for
 */
    Sender* getSender(Topic const& t) {
        std::lock_guard<std::mutex> lock(sendersLock_);
        auto sender = senders_.find(t);
        if ( sender != senders_.end()) {
            return sender->second.get();
        } else {
            std::lock_guard<std::mutex> slock(sendTransportEnginesLock_);
            for (auto i = 0u;
                i < sendTransportEngines_.size();
                ++i) {
                auto st = sendTransportEngines_[i];
                if (st->match(t)) {
                    auto newSender = new Sender(st, t);
                    senders_[t].reset(newSender);
                    return newSender;
                }
            }
        }

        return nullptr;
    }

/**
 * @brief This process is interested in a Topic
 * @details Normally the receiving transport covering this topic
 * needs to be created - not necessarily running - before calling this 
 * 
 * @param t Topic interested
 */
    void listenTo(Topic const& t) {
        std::lock_guard<std::mutex> tlock(recvTransportEnginesLock_);
        for (auto ptr : recvTransportEngines_) {
            ptr->listenTo(t);
        }
    }

/**
 * @brief undo the subscription
 * 
 * @param t Topic
 */
    void stopListenTo(Topic const& t) {
        std::lock_guard<std::mutex> tlock(recvTransportEnginesLock_);
        for (auto ptr : recvTransportEngines_) {
            ptr->stopListenTo(t);
        }
    }

private:
    NetContext() {
    }

    ~NetContext() {
    }


    std::vector<SendTransportEngine::ptr> sendTransportEngines_;
    std::mutex sendTransportEnginesLock_;

    std::vector<RecvTransport::ptr> recvTransportEngines_;
    std::mutex recvTransportEnginesLock_;

    std::map<Topic, Sender::ptr> senders_;
    uint32_t sendTransportRotationIndex_;
    std::mutex sendersLock_;    
};

}

}}

