#include "hmbdc/Copyright.hpp"
#pragma once 


#define NETMAP_WITH_LIBS
#include <net/netmap_user.h>


#include "hmbdc/app/netmap/Messages.hpp"
#include "hmbdc/app/Client.hpp"
#include "hmbdc/app/LoggerT.hpp" //log errors
#include "hmbdc/app/Config.hpp"
#include "hmbdc/comm/Topic.hpp"
#include "hmbdc/comm/eth/Misc.h"
#include "hmbdc/Compile.hpp"
#include "hmbdc/text/StringTrieSet.hpp"
#include "hmbdc/pattern/MonoLockFreeBuffer.hpp"
#include "hmbdc//Traits.hpp"


#include <boost/lexical_cast.hpp>
#include <memory>
#include <type_traits>


#include <netinet/ether.h>      /* ether_aton */
#include <linux/if_packet.h>    /* sockaddr_ll */
#include <sys/sysctl.h> /* sysctl */
#include <ifaddrs.h>    /* getifaddrs */
#include <poll.h>    

namespace hmbdc { namespace app { namespace netmap { 

using namespace hmbdc;
using namespace hmbdc::app;
using namespace hmbdc::comm;
using namespace hmbdc::comm::eth;
using namespace hmbdc::text;
/**
 * @brief power a netmap port receiving functions
 * @details this needs to be created using NetContext and start in an app::Context
 */
struct RecvTransport {
    using ptr = std::shared_ptr<RecvTransport>;
    virtual ~RecvTransport(){}
    /**
     * @brief a take all arbitrator (no arbitration at all)
     * @details it provides the default type for arbitration which does nothing
     * it also provides a template on writing a user defined arbitrator
     * 
     * @param h handle to retrieve what is inside of the message
     * @return always returns 1 here
     * In general: 1 keep the message; -1 to drop; 
     * 0 cannot decide, ask me later.
     */
    struct NoOpArb {
        int operator()(TransportMessageHeader const* h) {
            return 1; //always keep it
        }
    };
private:
    friend class NetContext; //only be used by NetContext
    virtual void listenTo(Topic const& t) = 0;
    virtual void stopListenTo(Topic const& t) = 0;
};

/**
 * @brief impl class, 
 * @details this needs to be created using NetContext and start in an app::Context
 * 
 * @tparam OutputBuffer type of buffer to hold resulting network messages
 * @tparam MsgArbitrator arbitrator to decide drop or keep messages, suited to arbitrate 
 * between different recv transports. By default, keeping all
 */
template <typename OutBuffer, typename MsgArbitrator = RecvTransport::NoOpArb>
struct RecvTransportEngineImpl 
: RecvTransport
, Client<RecvTransportEngineImpl<OutBuffer, MsgArbitrator>>
, MessageHandler<RecvTransportEngineImpl<OutBuffer, MsgArbitrator>, Subscribe, Unsubscribe> {
private:
    using MH = MessageHandler<RecvTransportEngineImpl<OutBuffer, MsgArbitrator>, Subscribe, Unsubscribe>;
    friend class NetContext; //only be created by NetContext

/**
 * @brief ctor
 * 
 * @param cfg specify the details of the mcast transport
 * @param outputBuffer holding the results
 * @param arb arbitrator instance to decide which messages to drop and keep; it supports either
 * raw netmap (udp) packet (BEFORE topic filtering) or hmbdc message (AFTER topic filtering) level 
 * arbitration depending on which one of 
 * int operator()(void* bytes, size_t len) or 
 * int operator()(TransportMessageHeader const* header) presents in the arb
 */    
    RecvTransportEngineImpl(Config const& config, OutBuffer& outBuffer
        , MsgArbitrator arb = NoOpArb())
    : config_(config)
    , buffer_(std::max(sizeof(Subscribe), sizeof(Unsubscribe))
        , config_.getExt<uint16_t>("cmdBufferSizePower2"))
    , outBuffer_(outBuffer)
    , maxMessageSize_(outBuffer.maxItemSize())
    , doChecksum_(config_.getExt<bool>("doChecksum"))
    , busyWait_(true)
    , pollWaitTimeMillisec_(0)
    , arb_(arb)
    , data_(nullptr) {
        config (hmbdcName_, "hmbdcName")
            (schedPolicy_, "schedPolicy")
            (schedPriority_, "schedPriority")
        ;

        struct nmreq baseNmd;
        bzero(&baseNmd, sizeof(baseNmd));
        baseNmd.nr_flags |= NR_ACCEPT_VNET_HDR;
        auto nmport = config_.getExt<std::string>("netmapPort");
        busyWait_ = config_.getExt<bool>("busyWait");
        if (!busyWait_) pollWaitTimeMillisec_ = config_.getExt<int>("pollWaitTimeMillisec");
        nmd_ = nm_open(nmport.c_str(), &baseNmd, config_.getExt<int>("nmOpenFlags"), NULL);
        if (!nmd_) {
            HMBDC_THROW(std::runtime_error, "cannot open " << nmport);
        }

        struct nmreq req;
        memset(&req, 0, sizeof(req));
        bcopy(nmd_->req.nr_name, req.nr_name, sizeof(req.nr_name));
        req.nr_version = NETMAP_API;
        req.nr_cmd = NETMAP_VNET_HDR_GET;
        int err = ioctl(nmd_->fd, NIOCREGIF, &req);
        if (err) {
            HMBDC_THROW(std::runtime_error, "Unable to get virtio-net header length");
        }
        virtHeader_ = req.nr_arg1;

        //setting up poll - might be useful or not
        pfd_.fd = nmd_->fd;
        pfd_.events = POLLIN;
        sleep(config_.getExt<int>("nmResetWaitSec"));
        //cleanup rings
        if (unlikely(ioctl(nmd_->fd, NIOCRXSYNC, NULL) < 0)) {
            HMBDC_THROW(std::runtime_error, "IO error");
        }
        for (int i = nmd_->first_rx_ring; i <= nmd_->last_rx_ring; i++) {
            struct netmap_ring * rxring = NETMAP_TXRING(nmd_->nifp, i);
            if (nm_ring_empty(rxring))
                continue;
            rxring->head = rxring->cur = rxring->tail;
        }
    }

    void listenTo(Topic const& t) override {
        buffer_.put(MessageWrap<Subscribe>{t});
    }

    void stopListenTo(Topic const& t) override {
        buffer_.put(MessageWrap<Unsubscribe>{t});
    }

public:

    ~RecvTransportEngineImpl() {
        if (nmd_) nm_close(nmd_);
    }

    /*virtual*/ 
    void invokedCb(uint16_t threadSerialNumber) __restrict__ override  {
        MonoLockFreeBuffer::iterator begin, end;
        auto n = buffer_.peek(begin, end);
        auto it = begin;
        while (it != end) {
            MH::handleMessage(*static_cast<MessageHead*>(*it++));
        }
        buffer_.wasteAfterPeek(begin, n);

        syncNetmap();
        for (int i = nmd_->first_rx_ring; i <= nmd_->last_rx_ring; i++) {
            struct netmap_ring * rxring = NETMAP_RXRING(nmd_->nifp, i);
            if (nm_ring_empty(rxring))
                continue;

            recvPackets(rxring);
        }
    }

    void handleMessageCb(Subscribe const& t) {
        subscriptions_.add(boost::lexical_cast<std::string>(t.topic));
    }

    void handleMessageCb(Unsubscribe const& t) {
        subscriptions_.erase(boost::lexical_cast<std::string>(t.topic));
    }

    void stoppedCb(std::exception const& e) override {
        HMBDC_LOG_C(e.what());
    };

    char const* hmbdcName() const { 
        return this->hmbdcName_.c_str();
    }

    std::tuple<char const*, int> schedSpec() const {
        return std::make_tuple(this->schedPolicy_.c_str(), this->schedPriority_);
    }

private:
    template <bool is_raw_arb>
    typename enable_if<is_raw_arb, int>::type applyRawArb(size_t len) {
        return arb_(data_, len);
    }

    template <bool is_raw_arb>
    typename enable_if<!is_raw_arb, int>::type applyRawArb(size_t) {
        return 1;
    }

    template <bool is_raw_arb>
    typename enable_if<!is_raw_arb, int>::type applyArb(TransportMessageHeader* h) {
        return arb_(h);
    }

    template <bool is_raw_arb>
    typename enable_if<is_raw_arb, int>::type applyArb(TransportMessageHeader* h) {
        return 1;
    }

    /**
     * @brief sync using busy wait or poll depending on config
     * @details it turns out busy wait performance is very poor when using vale
     * poll works mostly, but it works well only when an enough timeout is given
     * less than 10 milli wont work well
     */
    void syncNetmap() __restrict__ {
        if (likely(busyWait_)) {
            if (unlikely(ioctl(nmd_->fd, NIOCRXSYNC, NULL) < 0)) {
                HMBDC_THROW(std::runtime_error, "IO error");
            } else {
                return;
            }
        } else {
            auto res = poll(&pfd_, 1, pollWaitTimeMillisec_);
            if (unlikely( res < 0)) {
                HMBDC_THROW(std::runtime_error, "IO error errno=" << errno);
            } else {
                return;
            }
        }
    }

    void recvPackets(struct netmap_ring * __restrict__ ring) __restrict__ {
        auto cur = ring->cur;

        while(cur != ring->tail) {
            struct netmap_slot *slot = &ring->slot[cur];
            auto *buf = (uint8_t*)NETMAP_BUF(ring, slot->buf_idx);
            auto* p = reinterpret_cast<pkt*>(buf + virtHeader_ - sizeof(virt_header));

            if (p->ipv4.ip.ip_p == IPPROTO_UDP 
                && p->ipv4.ip.ip_off == htons(IP_DF)) {
                if (!data_) data_ = (uint8_t*)p->ipv4.body;

                using traits = 
                    pattern::function_traits<typename std::remove_reference<MsgArbitrator>::type>;
                using arg0 = typename traits::template arg<0>::type;
                bool const is_raw_arb = std::is_same<void*, typename std::decay<arg0>::type>::value;
                auto a = applyRawArb<is_raw_arb>(slot->len);
                if (unlikely(a == 0)) {
                    //keep data_ unchanged
                    ring->head = ring->cur = cur;
                    return;
                } else if (likely(a > 0)) {
                    while (data_ + sizeof(TransportMessageHeader) < buf + slot->len) {
                        auto header = reinterpret_cast<TransportMessageHeader*>(data_);
                        if (data_ + header->wireSize() <= buf + slot->len) {
                            if (subscriptions_.check(header->topic())) {
                                auto a = applyArb<is_raw_arb>(header);
                                if (likely(a > 0)) {
                                    auto l = std::min((size_t)header->messagePayloadLen(), maxMessageSize_);
                                    auto it = outBuffer_.claim();
                                    char* b = static_cast<char*>(*it);
                                    memcpy(b, header->payload(), l);
                                    outBuffer_.commit(it);
                                } else if (unlikely(a == 0)) {
                                    ring->head = ring->cur = cur;
                                    return;
                                } //else drop and move to next msg
                            }
                            data_ += header->wireSize();
                        } else {
                            break;
                        }
                    }
                } //else drop the packet
            }
            data_ = nullptr;
            cur = nm_ring_next(ring, cur);
        }
        
        ring->head = ring->cur = cur;
    }   

    static 
    bool verifyChecksum(pkt* __restrict__ packet, size_t payloadWireSize) {
        {
            auto tmp = packet->ipv4.ip.ip_sum;
            packet->ipv4.ip.ip_sum = 0;
            if (tmp != wrapsum(
                checksum(&packet->ipv4.ip, sizeof(packet->ipv4.ip), 0))) {
                return false;
            }
            packet->ipv4.ip.ip_sum = tmp;
        }
        
        {
            auto tmp = packet->ipv4.udp.check;
            packet->ipv4.udp.check = 0;

            auto udp = &packet->ipv4.udp;
            if (tmp != wrapsum(
                checksum(udp, sizeof(*udp), /* udp header */
                    checksum(packet->ipv4.body, payloadWireSize, /* udp payload */ 
                        checksum(&packet->ipv4.ip.ip_src, 2 * sizeof(packet->ipv4.ip.ip_src), /* pseudo header */
                            IPPROTO_UDP + (u_int32_t)ntohs(udp->len)))))) {
                return false;
            }
            packet->ipv4.udp.check = tmp;
            return true;
        }
    }
    Config config_;
    string hmbdcName_;
    string schedPolicy_;
    int schedPriority_;

    hmbdc::pattern::MonoLockFreeBuffer buffer_;
    OutBuffer& __restrict__ outBuffer_;
    size_t maxMessageSize_;
    StringTrieSet subscriptions_;


    struct nm_desc *nmd_;
    int virtHeader_; //v hdr len
    bool doChecksum_;
    struct pollfd pfd_;
    bool busyWait_;
    int pollWaitTimeMillisec_;
    MsgArbitrator arb_;
    uint8_t* data_;
};

}}}

