#include "hmbdc/Copyright.hpp"
#pragma once 


#define NETMAP_WITH_LIBS
#include <net/netmap_user.h>

#include "hmbdc/app/netmap/Messages.hpp"
#include "hmbdc/app/Client.hpp"
#include "hmbdc/app/LoggerT.hpp" //log errors
#include "hmbdc/pattern/MonoLockFreeBuffer.hpp"
#include "hmbdc/comm/Topic.hpp"
#include "hmbdc/comm/eth/Misc.h"
#include "hmbdc/time/Rater.hpp"
#include "hmbdc/numeric/BitMath.hpp"

#include "hmbdc/app/Config.hpp"
#include "hmbdc/Compile.hpp"


#include <boost/regex.hpp>
#include <memory>
#include <type_traits>


#include <netinet/ether.h>      /* ether_aton */
#include <linux/if_packet.h>    /* sockaddr_ll */
#include <sys/sysctl.h> /* sysctl */
#include <ifaddrs.h>    /* getifaddrs */

#include <poll.h>    

namespace hmbdc { namespace app { namespace netmap { 

using namespace hmbdc;
using namespace hmbdc::time;
using namespace hmbdc::comm;
using namespace hmbdc::comm::eth;

/**
 * @brief power a netmap port sending functions
 * @details this needs to be created using NetContext and start in an app::Context
 * 
 */
struct SendTransportEngine 
: Client<SendTransportEngine> {
    using ptr = std::shared_ptr<SendTransportEngine>;

private:
    uint16_t
    outBufferSizePower2() {
        auto res = config_.getExt<uint16_t>("outBufferSizePower2");
        if (res) {
            return res;
        }
        res =hmbdc::numeric::log2Upper(16ul * 1024ul / (8ul + maxMessageSize_));
        HMBDC_LOG_N("auto set --outBufferSizePower2=", res);
        return res;
    }

    friend class NetContext; //only be created by NetContext
    SendTransportEngine(Config const& cfg, size_t maxMessageSize) 
    : config_(cfg)
    , topic_(cfg.getExt<string>("topicRegex"))
    , topicRegex_(topic_) 
    , maxMessageSize_(maxMessageSize)
    , nmd_(nullptr)
    , buffer_(maxMessageSize + sizeof(TransportMessageHeader) + sizeof(Topic)
        , outBufferSizePower2())
    , virtHeader_(0)
    , srcEthAddr_{0}
    , dstEthAddr_{0}
    , doChecksum_(config_.getExt<bool>("doChecksum"))
    , rater_(Duration::seconds(1u)
        , config_.getExt<size_t>("sendBytesPerSec")
        , config_.getExt<size_t>("sendBytesBurst")
        , config_.getExt<size_t>("sendBytesBurst") != 0ul //no rate control by default
    ) 
    , maxSendBatch_(cfg.getExt<size_t>("maxSendBatch")) 
    , mtu_(config_.getExt<size_t>("mtu")) {
        mtu_ -= (8u + 20u);  // 8bytes udp header and 20bytes ip header
        cfg (hmbdcName_, "hmbdcName")
            (schedPolicy_, "schedPolicy")
            (schedPriority_, "schedPriority")
        ;

        memcpy(&srcEthAddr_, ether_aton(config_.getExt<std::string>("srcEthAddr").c_str())
            , sizeof(srcEthAddr_));
        memcpy(&dstEthAddr_, ether_aton(config_.getExt<std::string>("dstEthAddr").c_str())
            , sizeof(dstEthAddr_));
        getMacAddresses();

        struct nmreq baseNmd;
        bzero(&baseNmd, sizeof(baseNmd));
        baseNmd.nr_flags |= NR_ACCEPT_VNET_HDR;

        auto nmport = cfg.getExt<std::string>("netmapPort");
        nmd_ = nm_open(nmport.c_str(), &baseNmd
            , cfg.getExt<uint64_t>("nmOpenFlags"), NULL);
        if (!nmd_) {
            HMBDC_THROW(std::runtime_error, "cannot open " << nmport);
        }

        struct nmreq req;
        memset(&req, 0, sizeof(req));
        bcopy(nmd_->req.nr_name, req.nr_name, sizeof(req.nr_name));
        req.nr_version = NETMAP_API;
        req.nr_cmd = NETMAP_VNET_HDR_GET;
        int err = ioctl(nmd_->fd, NIOCREGIF, &req);
        if (err) {
            HMBDC_THROW(std::runtime_error, "Unable to get virtio-net header length");
        }
        virtHeader_ = req.nr_arg1;

        initializePacket(&precalculatedPacketHead_
            , config_.getExt<uint16_t>("ttl")
            , config_.getExt<string>("srcIp")
            , config_.getExt<string>("dstIp")
            , srcEthAddr_
            , dstEthAddr_
            , config_.getExt<uint16_t>("srcPort")
            , config_.getExt<uint16_t>("dstPort")
        );
        sleep(config_.getExt<int>("nmResetWaitSec"));
        //cleanup rings
        if (unlikely(ioctl(nmd_->fd, NIOCTXSYNC, NULL) < 0)) {
            HMBDC_THROW(std::runtime_error, "IO error");
        }
        for (int i = nmd_->first_tx_ring; i <= nmd_->last_tx_ring; i++) {
            struct netmap_ring * txring = NETMAP_TXRING(nmd_->nifp, i);
            if (nm_ring_empty(txring))
                continue;
            txring->head = txring->cur = txring->tail;
        }
    }

public:

    ~SendTransportEngine() {
        nm_close(nmd_);
    }

    /*virtual*/ bool droppedCb() override {
        buffer_.reset();
        return true;
    }

    /*virtual*/ 
    void invokedCb(uint16_t threadSerialNumber) __restrict__ override  {
        if (unlikely(ioctl(nmd_->fd, NIOCTXSYNC, NULL) < 0)) {
            HMBDC_THROW(std::runtime_error, "IO error");
        }

        for (int i = nmd_->first_tx_ring; i <= nmd_->last_tx_ring; i++) {
            struct netmap_ring * txring = NETMAP_TXRING(nmd_->nifp, i);
            if (nm_ring_empty(txring))
                continue;

            sendPackets(txring);
        }
    }

    void stoppedCb(std::exception const& e) override {
        HMBDC_LOG_C(e.what());
    };

    char const* hmbdcName() const { 
        return this->hmbdcName_.c_str();
    }

    std::tuple<char const*, int> schedSpec() const {
        return std::make_tuple(this->schedPolicy_.c_str(), this->schedPriority_);
    }

private:
    bool match(Topic const& t) const {
        return boost::regex_match(t.c_str(), topicRegex_);
    }
    friend class Sender; 
    template <typename... Messages>
    void queue(Topic const& t, Messages&&... msgs) __restrict__ {
        auto n = sizeof...(msgs);
        auto it = buffer_.claim(n);
        queue(it, t, std::forward<Messages>(msgs)...);
        buffer_.commit(it, n);
    }

    template <typename... Messages>
    bool tryQueue(Topic const& t, Messages&&... msgs) __restrict__ {
        auto n = sizeof...(msgs);
        auto it = buffer_.tryClaim(n);
        if (it) {
            queue(it, t, std::forward<Messages>(msgs)...);
            buffer_.commit(it, n);
            return true;
        }
        return false;
    }

    template <typename M, typename... Messages>
    void queue(hmbdc::pattern::MonoLockFreeBuffer::iterator it
        , Topic const& t, M&& msg, Messages&&... msgs) {
        using Message = typename std::remove_reference<M>::type;
        if (unlikely(sizeof(Message) > maxMessageSize_)) {
            HMBDC_THROW(std::out_of_range, "maxMessageSize too small to hold a message when constructing SendTransportEngine");        
        }
        auto s = *it;
        TransportMessageHeader::copyTo(s, t, std::forward<M>(msg));
        queue(++it, t, std::forward<M>(msgs)...);
    }

    template <typename Message, typename ... Args>
    void queueInPlace(Topic const& t, Args&&... args) __restrict__ {
        if (unlikely(sizeof(Message) > maxMessageSize_)) {
            HMBDC_THROW(std::out_of_range, "maxMessageSize too small to hold a message when constructing SendTransportEngine");        
        }
        auto s = buffer_.claim();
        TransportMessageHeader::copyToInPlace<Message>(*s, t, std::forward<Args>(args)...);
        buffer_.commit(s);
    }

    void queueBytes(Topic const& t, uint16_t tag, void const* bytes, size_t len) {
        if (unlikely(len > maxMessageSize_)) {
            HMBDC_THROW(std::out_of_range, "maxMessageSize too small to hold a message when constructing SendTransportEngine");        
        }
        auto s = buffer_.claim();
        TransportMessageHeader::copyTo(*s, t, tag, bytes, len);
        buffer_.commit(s);
    }

    void queue(hmbdc::pattern::MonoLockFreeBuffer::iterator it
        , Topic const& t) {}

    void sendPackets(struct netmap_ring * __restrict__ ring) __restrict__ {
        uint32_t cur = ring->cur;
        if (unlikely(cur == ring->tail)) return;
        MonoLockFreeBuffer::iterator begin, end;
        if (unlikely(!(buffer_.peek(begin, end, maxSendBatch_)))) {
            return;
        }
        bool slotInited = false;
        auto it = begin;
        auto batch = maxSendBatch_;
        struct netmap_slot *slot = &ring->slot[cur];
        uint32_t slotLen = 0;
        char *p = NETMAP_BUF(ring, slot->buf_idx);
        pkt* currentPktPtr = (pkt*)(p + virtHeader_ - sizeof(virt_header));
        uint16_t slotLenMax = min(mtu_, (uint16_t)ring->nr_buf_size);
        for (; it != end;) {
            auto th = reinterpret_cast<TransportMessageHeader*>(*it);
            if (rater_.check(th->wireSize())) {
                if (!slotInited) {
                    auto wireSize = (uint16_t)(
                        sizeof(ether_header) + sizeof(ip) + sizeof(udphdr) + virtHeader_
                    );
                    memcpy(p, ((char*)&precalculatedPacketHead_) + sizeof(virt_header) - virtHeader_
                        , wireSize);
                    slotLen = wireSize;
                    slotInited = true;
                }
                auto wireSize = th->wireSize();
                if (slotLen + wireSize <= slotLenMax) {
                    memcpy(p + slotLen, th, (int)wireSize);
                    slotLen += wireSize;
                    rater_.commit();
                    batch--;
                    ++it;
                } else {
                    batch = 0; //this batch is done
                }
                if (!batch) {
                    slot->len = slotLen;
                    size_t wireSizeExcludingHead = slotLen 
                        - (sizeof(ether_header) + sizeof(ip) + sizeof(udphdr) + virtHeader_);
                    updatePacket(currentPktPtr, wireSizeExcludingHead, doChecksum_);
                    cur = nm_ring_next(ring, cur);
                    slotLen = 0;
                    if (cur == ring->tail) break;
                    slot = &ring->slot[cur];
                    p = NETMAP_BUF(ring, slot->buf_idx);
                    currentPktPtr = (pkt*)(p + virtHeader_ - sizeof(virt_header));
                    batch = maxSendBatch_;
                    slotInited = false;
                }
            } else {
                break;
            }
        }
        if (slotLen) {
            slot->len = slotLen;
            size_t wireSizeExcludingHead = slotLen 
                - (sizeof(ether_header) + sizeof(ip) + sizeof(udphdr) + virtHeader_);
            updatePacket(currentPktPtr, wireSizeExcludingHead, doChecksum_);
            cur = nm_ring_next(ring, cur);
        }

        ring->head = ring->cur = cur;
        buffer_.wasteAfterPeek(begin, it - begin, true);
    }

    void getMacAddresses() {
        auto nmport = config_.getExt<std::string>("netmapPort");

        if (strncmp(nmport.c_str(), "vale", 4) == 0) return;
        
        if (nmport.find_first_of(":") == std::string::npos) {
            HMBDC_THROW(std::runtime_error
                , "wrong netmapPort format (examples: netmap:eth0, netmap:eth0-0)");
        }
        auto iface = nmport.substr(nmport.find_first_of(":")); 
        iface = iface.substr(1, iface.find_first_of("-^") - 1);


        struct ifaddrs *ifaphead, *ifap;
        int l = sizeof(ifap->ifa_name);

        if (getifaddrs(&ifaphead) != 0) {
            HMBDC_THROW(std::runtime_error, "getifaddrs failed for" << iface);
        }
        for (ifap = ifaphead; ifap; ifap = ifap->ifa_next) {
            struct sockaddr_ll *sll =
                (struct sockaddr_ll *)ifap->ifa_addr;
            uint8_t *mac;

            if (!sll || sll->sll_family != AF_PACKET)
                continue;
            if (strncmp(ifap->ifa_name, iface.c_str(), l) != 0)
                continue;
            mac = (uint8_t *)(sll->sll_addr);

            char srcEthAddrStr[20];
            sprintf(srcEthAddrStr, "%02x:%02x:%02x:%02x:%02x:%02x",
                mac[0], mac[1], mac[2],
                mac[3], mac[4], mac[5]);
            memcpy(&srcEthAddr_, ether_aton(srcEthAddrStr), sizeof(srcEthAddr_)); //6 bytes
            break;
        }
        freeifaddrs(ifaphead);
        if (!ifap) {
            HMBDC_THROW(std::runtime_error, "no local interface named " << iface);
        }
    }

    static
    void initializePacket(struct pkt *pkt, int ttl, std::string srcIpStr, std::string dstIpStr
        , ether_addr srcEthAddr, ether_addr dstEthAddr, uint16_t srcPort, uint16_t dstPort) {
        struct ether_header *eh;
        struct ip *ip;
        struct udphdr *udp;
        uint32_t a, b, c, d;
        sscanf(srcIpStr.c_str(), "%d.%d.%d.%d", &a, &b, &c, &d);
        auto srcIp = (a << 24u) + (b << 16u) + (c << 8u) + d;
        sscanf(dstIpStr.c_str(), "%d.%d.%d.%d", &a, &b, &c, &d);
        auto dstIp = (a << 24u) + (b << 16u) + (c << 8u) + d;

        /* prepare the headers */
        eh = &pkt->eh;
        bcopy(&srcEthAddr, eh->ether_shost, 6);
        bcopy(&dstEthAddr, eh->ether_dhost, 6);

        eh->ether_type = htons(ETHERTYPE_IP);
        ip = &pkt->ipv4.ip;
        udp = &pkt->ipv4.udp;
        ip->ip_v = IPVERSION;
        ip->ip_hl = sizeof(*ip) >> 2;
        ip->ip_id = 0;
        ip->ip_tos = IPTOS_LOWDELAY;
        ip->ip_len = 0; //zero so chksum can happen in ip_sum
        ip->ip_id = 0;
        ip->ip_off = htons(IP_DF); /* Don't fragment */
        ip->ip_ttl = ttl;
        ip->ip_p = IPPROTO_UDP;
        ip->ip_dst.s_addr = htonl(dstIp); 
        ip->ip_src.s_addr = htonl(srcIp); 
        ip->ip_sum = 0;
        ip->ip_len = sizeof(*ip) + sizeof(udphdr); //ip->ip_len is unknown, put known part
        udp->source = htons(srcPort);
        udp->dest = htons(dstPort);
        udp->len = sizeof(udphdr); //put known part
        udp->check = 0;
        
        bzero(&pkt->vh, sizeof(pkt->vh));
    }

    static 
    void updatePacket(struct pkt *packet, size_t payloadWireSize, bool doChecksum = true) {
        packet->ipv4.ip.ip_len += payloadWireSize; //already has sizeof(ip) + sizeof(udphdr);
        packet->ipv4.ip.ip_len = ntohs(packet->ipv4.ip.ip_len);
        if (doChecksum) {
            packet->ipv4.ip.ip_sum = wrapsum(checksum(&packet->ipv4.ip, sizeof(packet->ipv4.ip), 0));
        }
        
        packet->ipv4.udp.len += payloadWireSize;
        packet->ipv4.udp.len = htons(packet->ipv4.udp.len);
        if (doChecksum) {
            auto udp = &packet->ipv4.udp;
            packet->ipv4.udp.check = wrapsum(
                checksum(udp, sizeof(*udp), /* udp header */
                    checksum(packet->ipv4.body, payloadWireSize, /* udp payload */ 
                        checksum(&packet->ipv4.ip.ip_src, 2 * sizeof(packet->ipv4.ip.ip_src), /* pseudo header */
                            IPPROTO_UDP + (u_int32_t)ntohs(udp->len)))));
        }
    }

    
    Config const config_;
    string hmbdcName_;
    string schedPolicy_;
    int schedPriority_;
    std::string topic_;
    boost::regex topicRegex_;
    size_t maxMessageSize_;


    struct nm_desc *nmd_;
    hmbdc::pattern::MonoLockFreeBuffer buffer_;
    int virtHeader_; //v hdr len


    ether_addr srcEthAddr_;
    ether_addr dstEthAddr_;
    pkt precalculatedPacketHead_;
    bool doChecksum_;
    Rater rater_;
    size_t maxSendBatch_;
    uint16_t mtu_;
};

}}}
