#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/Config.hpp"
#include "hmbdc/app/tcpcast/Messages.hpp"
#include "hmbdc/app/tcpcast/Sender.hpp"
#include "hmbdc/app/tcpcast/SendTransportEngine.hpp"
#include "hmbdc/app/tcpcast/RecvTransportEngine.hpp"
#include "hmbdc/app/tcpcast/DefaultUserConfig.hpp"

#include "hmbdc/comm/Topic.hpp"
#include "hmbdc/pattern/GuardedSingleton.hpp"

#include <boost/regex.hpp>

#include <vector>
#include <mutex>
#include <memory>

namespace hmbdc { namespace app { namespace tcpcast {

/**
 * @brief a singleton that holding tcpcast resources
 * @details it manages transport engines
 * 
 * see @example perf-tcpcast.cpp @example chat.cpp for usage.
 * 
 */
struct NetContext 
: hmbdc::pattern::GuardedSingleton<NetContext> {

    friend class hmbdc::pattern::SingletonGuardian<NetContext>;
    using Sender = hmbdc::app::tcpcast::Sender;

/**
 * @brief construct a send transport enngine (and remember it)
 * @details After this, user is responsible to get it started within a hmbdc Context,
 * if running in Context pool, it needs to be pinned on a single pool thread, CANNOT span
 * more than one thread, otherwise the transport is not functioing/running. 
 * Don't create the same thing twice.
 * 
 * @param cfgIn jason specifing the transport - see example, perf-tcpcast.cpp and DefaultUserConfig.hpp
 * @param maxMessageSize max messafe size in bytes to be sent
 * @param minRecvToStart start send when there are that many recipients (processes) online, otherwise
 * hold the message in buffer
 * @return a pointer to the Engine
 */
    SendTransportEngine* createSendTransportEngine(
        Config const& cfgIn
        , size_t maxMessageSize
        , size_t minRecvToStart = 1u) {
        Config dft(DefaultUserConfig, "tx");
        Config cfg(cfgIn);
        cfg.setDefaultUserConfig(dft);

        std::lock_guard<std::mutex> tlock(sendTransportEnginesLock_);
        auto res = new SendTransportEngine(
            cfg, maxMessageSize, minRecvToStart);
        sendTransports_.emplace_back(res);
        return res;
    }

/**
 * @brief same as above but provide an unified interface - not preferred
 * @return a pointer to the Engine - don't delete it
 */
    SendTransportEngine* createSendTransportEngineTuply(Config const& cfg
        , size_t maxMessageSize
        , tuple<size_t> args) {
        return createSendTransportEngine(cfg, maxMessageSize, get<0>(args));
    }

/**
 * @brief construct a send transport and remember it
 * @details After this, user is respoonsible to get it started within a hmbdc Context,
 * if running in Context pool, it needs to pin at a single pool thread, CANNOT span
 * more than one thread
 * otherwise the transport is not functioing/running. Don't create the same thing twice
 * 
 * @param cfgIn jason specifing the transport - see example, perf-tcpcast.cpp and DefaultUserConfig.hpp
 * @param buffer buffer that recv messages go, normally the one returned by app::Context::buffer()
 * @param arb optonally an arbitrator to decide which messages to keep and drop
 * if arb is an rvalue, it is passed in value, if an lvalue, passed in as reference;
 * it supports ONLY hmbdc message level (AFTER topic filtering) arbitration if
 * int operator()(TransportMessageHeader const* header) presents in the arb passed in.
 * (NO packet level since it is tcp)
 * @return a pointer to the Engine
 */
    template <typename Buffer, typename MsgArbitrator = RecvTransport::NoOpArb>
    auto createRecvTransportEngine(Config const& cfgIn
        , Buffer& buffer
        , MsgArbitrator&& arb = RecvTransport::NoOpArb()) {
        Config dft(DefaultUserConfig, "rx");
        Config cfg(cfgIn);
        cfg.setDefaultUserConfig(dft);

        std::lock_guard<std::mutex> tlock(recvTransportsLock_);
        auto res = new RecvTransportEngine<Buffer, MsgArbitrator>(
            cfg, buffer, forward<MsgArbitrator>(arb));
        recvTransports_.emplace_back(res);
        return res;
    }

/**
 * @brief same as above but to provide a unified interface - not preferred 
 * @details use forward_as_tuple to make the tuple passed in
 * @return a pointer to the Engine
 */
    template <typename Buffer, typename ArgsTuple>
    auto createRecvTransportEngineTuply(
        Config const& cfg
        , Buffer& buffer
        , ArgsTuple&& args) {
        return createRecvTransportEngine(cfg
            , buffer
            , get<0>(args)
        );
    }

/**
 * @brief get (or create for the first time) a Sender - whose function is to send messages on 
 * its associated Topic
 * @details this operation typically might be slow, so caching the return value is recommended. 
 * 
 * @param t - the Topic that the Sender is for
 */
    Sender* getSender(Topic const& t) {
        std::lock_guard<std::mutex> lock(sendersLock_);
        auto sender = senders_.find(t);
        if ( sender != senders_.end()) {
            return sender->second.get();
        } else {
            std::lock_guard<std::mutex> slock(sendTransportEnginesLock_);
            for (auto i = 0u;
                i < sendTransports_.size();
                ++i) {
                auto st = sendTransports_[i];
                if (st->match(t)) {
                    auto newSender = new Sender(st, t);
                    senders_[t].reset(newSender);
                    return newSender;
                }
            }
        }

        return nullptr;
    }

/**
 * @brief This process is interested in a Topic
 * @details Normally the receiving transport covering this topic
 * needs to be created - not necessarily running - before calling this 
 * 
 * @param t Topic interested
 */
    void listenTo(Topic const& t) {
        std::lock_guard<std::mutex> tlock(recvTransportsLock_);
        for (auto ptr : recvTransports_) {
            ptr->listenTo(t);
        }
    }

/**
 * @brief undo the subscription
 * 
 * @param t Topic
 */
    void stopListenTo(Topic const& t) {
        std::lock_guard<std::mutex> tlock(recvTransportsLock_);
        for (auto ptr : recvTransports_) {
            ptr->stopListenTo(t);
        }
    }

private:
    NetContext() {
    }

    ~NetContext() {
    }


    std::vector<SendTransport::ptr> sendTransports_;
    std::mutex sendTransportEnginesLock_;

    std::vector<RecvTransport::ptr> recvTransports_;
    std::mutex recvTransportsLock_;

    std::map<Topic, Sender::ptr> senders_;
    std::mutex sendersLock_;    
};

} //tcpcast

}}

