#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/LoggerT.hpp"
#include "hmbdc/app/tcpcast/Transport.hpp"
#include "hmbdc/app/tcpcast/Messages.hpp"
#include "hmbdc/time/Time.hpp"
#include "hmbdc/os/DownloadFile.hpp"
#include "hmbdc/os/DownloadMemory.hpp"
#include "hmbdc/comm/inet/Misc.hpp"

#include "hmbdc/app/tcpcast/RecvSession.hpp"

#include <boost/asio.hpp>
#include <boost/lexical_cast.hpp>

#include <functional>
#include <memory>
#include <utility>

#include <iostream>

namespace hmbdc { namespace app { namespace tcpcast {

using namespace boost::asio;
using boost::asio::ip::tcp;
using namespace std;

template <typename OutputBuffer, typename MsgArbitrator>
struct RecvSession 
: enable_shared_from_this<RecvSession<OutputBuffer, MsgArbitrator>> {
    using ptr = shared_ptr<RecvSession<OutputBuffer, MsgArbitrator>>;
    using CleanupFunc = std::function<void()>;
    RecvSession(Config const& config
        , StringTrieSet const& subscriptions
        , boost::regex const& topicRegex
        , CleanupFunc cleanupFunc
        , io_service& ios
        , OutputBuffer& outputBuffer
        , MsgArbitrator& arb)
    : config_(config)
    , subscriptions_(subscriptions)
    , topicRegex_(topicRegex)
    , cleanupFunc_(cleanupFunc)
    , socket_(ios)
    , arb_(arb)
    , outputBuffer_(outputBuffer)
    , bufSize_(config.getExt<size_t>("maxTcpReadBytes"))
    , buf_((char*)memalign(SMP_CACHE_BYTES, bufSize_))
    , bufCur_(buf_)
    , filledLen_(0)
    , doReadTimer_(ios)
    , currTransportHeadFlag_(0)
    , downloadDir_(config.getExt<string>("downloadDir"))
    , allowRecvMemoryAttachment_(config.getExt<bool>("allowRecvMemoryAttachment")) {
    }

    ~RecvSession() {
        free(buf_);
        HMBDC_LOG_N("RecvSession retired: ", id());
    }

    void start(tcp::resolver::iterator endpointIt) {
        auto self(this->shared_from_this());
        async_connect(socket_, endpointIt,
            [this, self](boost::system::error_code ec, tcp::resolver::iterator) {
                if (!ec) {
                    id_ = boost::lexical_cast<string>(socket_.remote_endpoint());
                    strncpy(sessionStarted_.payload.ip, id_.c_str()
                        , sizeof(sessionStarted_.payload.ip));
                    sessionStarted_.payload.ip[sizeof(sessionStarted_.payload.ip) - 1] = 0;
                    strncpy(sessionDropped_.payload.ip, id_.c_str()
                        , sizeof(sessionDropped_.payload.ip));
                    sessionDropped_.payload.ip[sizeof(sessionDropped_.payload.ip) - 1] = 0;

                    auto sz = config_.getExt<int>("tcpRecvBufferBytes");
                    if (sz) {
                        socket_base::receive_buffer_size option(sz);
                        boost::system::error_code ec;
                        socket_.set_option(option);
                    }

                    socket_base::receive_buffer_size option;
                    socket_.get_option(option);
                    if (sz == 0 || sz >= option.value()) {
                        // HMBDC_LOG_N("RecvSession receive buffer size: ", option.value());
                    } else {
                        HMBDC_LOG_C("set tcpcast RecvSession receive buffer size unsuccessful, want "
                            , sz, " actual: ", option.value()
                            , " resulting higher receiver dropping possibility, check OS limits!");
                    }

                    doRead(); //async
                    socket_.non_blocking(true);
                    sendSubscriptions(); //sync
                    outputBuffer_.put(sessionStarted_);
                }
            }
        );
    }

    void sendSubscribe(Topic const& t) {
        if (unlikely(boost::regex_match(t.c_str(), topicRegex_))) {
            char buf[70];
            buf[0]='+';
            auto l = t.copyTo(buf + 1);
            buf[l + 1] = '\t';
            if (l + 2 != socket_.send(buffer(buf, l + 2))) {
                HMBDC_LOG_C("error when sending subscribe to ", id());
            }
        }
    }


    void sendUnsubscribe(Topic const& t) {
        if (unlikely(boost::regex_match(t.c_str(), topicRegex_))) {
            char buf[70];
            buf[0]='-';
            auto l = t.copyTo(buf + 1);
            buf[l + 1] = '\t';
            if (l + 2 != socket_.send(buffer(buf, l + 2))) {
                HMBDC_LOG_C("error when sending unsubscribe to ", id());
            }
        }
    }

    void heartbeat() {
        if (!socket_.is_open()) return;
        char const* hb = "+\t"; 
        if (socket_.send(buffer(hb, 2)) != 2) {
                HMBDC_LOG_C("error when sending heartbeat to ", id());
        }
    }

    char const* id() const {
        return id_.c_str();
    }

private:
    void sendSubscriptions() {
        ostringstream oss;
        for (auto it = subscriptions_.begin(); it != subscriptions_.end(); ++it) {
            auto const& t = it->first;
            if (boost::regex_match(t.c_str(), topicRegex_)) {
                oss << '+' << t;
                if (it->second) {
                    oss << '*';
                }
                oss << '\t';
            }
        }
        auto s = oss.str() + "+\t"; //mark all sent with "+\t"
        auto sz = socket_.send(buffer(s.c_str(), s.size()));
        if (sz != s.size()) {
            HMBDC_LOG_C("error when sending subscriptions to ", id());
        }


    }
    void doRead() {
        if (unlikely(currTransportHeadFlag_ == hasFileAttachment::flag)) {
            auto s = fileAttachment_.write(bufCur_, filledLen_);
            if (fileAttachment_.writeDone()) {
                fileAttachment_.close();
                currTransportHeadFlag_ = 0;
                outputBuffer_.commit(it_);
            }
            bufCur_ += s;
            filledLen_ -= s;
        } else if (unlikely(currTransportHeadFlag_ == hasMemoryAttachment::flag)) {
            auto s = memoryAttachment_.write(bufCur_, filledLen_);
            if (memoryAttachment_.writeDone()) {
                memoryAttachment_.close();
                currTransportHeadFlag_ = 0;
                outputBuffer_.commit(it_);
            }
            bufCur_ += s;
            filledLen_ -= s;
        }
        while (currTransportHeadFlag_ == 0 
            && filledLen_ >= sizeof(TransportMessageHeader)) {
            auto h = reinterpret_cast<TransportMessageHeader*>(bufCur_);
            auto wireSize = h->wireSize();
            if (likely(filledLen_ >= wireSize)) {
                if (likely(subscriptions_.check(h->topic()))) {
                    auto a = arb_(h);
                    if (likely(a > 0)) {
                        it_ = outputBuffer_.claim();
                        char* b = static_cast<char*>(*it_);
                        auto l = std::min<size_t>(outputBuffer_.maxItemSize(), h->messagePayloadLen);
                        memcpy(b, h->payload(), l);
                        currTransportHeadFlag_ = h->flag;
                        if (!currTransportHeadFlag_) {
                            outputBuffer_.commit(it_);
                        } else if (currTransportHeadFlag_ == hasMemoryAttachment::flag) {
                            auto& att = 
                                static_cast<app::MessageWrap<hasMemoryAttachment>*>(*it_)->payload;
                            if (allowRecvMemoryAttachment_) {
                                att.attachment = memoryAttachment_.open(att.len);
                            }
                        } else if (currTransportHeadFlag_ == hasFileAttachment::flag) {
                            auto& att = 
                                static_cast<app::MessageWrap<hasFileAttachment>*>(*it_)->payload;
                            if (downloadDir_.size()) {
                                fileAttachment_.open(downloadDir_.c_str(), att.file, att.len);
                            }
                            if (!fileAttachment_) {
                                att.attachment = nullptr;
                            }
                        }
                    } else if (unlikely(a == 0)) {
                        //cannot decide now - keep bufCur_ filledLen_ unchanged 
                        //try it later, use a timer since there might be no more bytes arriving
                        doReadTimer_.expires_from_now(boost::posix_time::microseconds(0));

                        auto self(this->shared_from_this());
                        doReadTimer_.async_wait(
                            [this, self](boost::system::error_code ec) {
                                if (!ec) doRead();
                            }
                        );
                        return;
                    } //else move to next msg
                }
                bufCur_ += wireSize;
                //memmove(buf_, buf_ + h->wireSize(), filledLen_ - h->wireSize());
                filledLen_ -= wireSize;
            } else {
                break;
            }
        }
        memmove(buf_, bufCur_, filledLen_);
        bufCur_ = buf_;

        auto self(this->shared_from_this());
        socket_.async_read_some(boost::asio::buffer(buf_ + filledLen_, bufSize_ - filledLen_),
            [this, self](boost::system::error_code ec, std::size_t length) {
                if (!ec) {
                    filledLen_ += length;
                    doRead();
                } else {
                    HMBDC_LOG_C(id(), " receiving error = ", ec);
                    socket_.cancel();
                    cleanupFunc_();
                    outputBuffer_.put(sessionDropped_);
                }
        });
    }

    Config const& config_;
    StringTrieSet const& subscriptions_;
    boost::regex topicRegex_;
    std::function<void()> cleanupFunc_;
    tcp::socket socket_;
    MsgArbitrator& arb_;
    OutputBuffer& outputBuffer_;
    string id_;
    MessageWrap<SessionStarted> sessionStarted_;
    MessageWrap<SessionDropped> sessionDropped_;
    size_t const bufSize_;
    char* buf_;
    char* bufCur_;
    size_t filledLen_;
    deadline_timer doReadTimer_;
    uint8_t currTransportHeadFlag_;
    typename OutputBuffer::iterator it_;
    string downloadDir_;
    os::DownloadFile fileAttachment_;
    bool allowRecvMemoryAttachment_;
    os::DownloadMemory memoryAttachment_;
};

}}}
