#include "hmbdc/app/LoggerT.hpp"
#include "hmbdc/app/Client.hpp"
#include "hmbdc/app/Context.hpp"
#include "hmbdc/text/Misc.h"
#include "hmbdc/numeric/BitMath.hpp"
#include "hmbdc/os/Signals.hpp"

#include <iostream>
#include <memory>
#include <unistd.h>
#include <boost/program_options.hpp>

namespace hmbdc { namespace app { namespace utils {

using namespace hmbdc;
using namespace hmbdc::app;
using namespace tested_module;
using namespace hmbdc::time;
using namespace hmbdc::numeric;
using namespace boost::property_tree;

using SendCtx = hmbdc::app::Context<24>;

struct Message
: hasTag<1001> {
    Message() 
    : seq(seq_s++)
    {}
    size_t seq;
    friend std::ostream& operator << (std::ostream& os, Message const& r) {
        os << "Message " << ' ' << r.seq;
        return os;
    }
    char padding[512 - sizeof(seq)];

    static uint64_t seq_s;
} __attribute__((__packed__));

uint64_t Message::seq_s = 0ul;

struct SenderClient 
: Client<SenderClient>
, TimerManager
, ReoccuringTimer {
    SenderClient(SendCtx& ctx, Topic const topic, size_t msgSize)
    : ReoccuringTimer(Duration::seconds(1u))
    , ctx_(ctx)
    , msgSize_(msgSize)
    , periodicMsgCount_(0) {
        sender_ = NetContext::instance().getSender(topic);
        setCallback(
            [this](TimerManager& tm, SysTime const& now) {
            report();
            }
        );
        schedule(SysTime::now(), *this);
    }

    char const* hmbdcName() const { 
        return "perf-tx"; 
    }
    
    /*virtual*/ 
    void invokedCb(uint16_t) __restrict__ override {
        //send as fast as possible since the engine has rate control
        Message m;
        sender_->send(m, msgSize_);
        periodicMsgCount_++;
    }

    void stoppedCb(std::exception const& e){
        std::cout << (e.what()) << std::endl;
    };

    void report() {
        HMBDC_LOG_n("msgSize=", msgSize_, " mps=", periodicMsgCount_, " totalMsgCount=", Message::seq_s);
        periodicMsgCount_ = 0u;
    }

    SendCtx& ctx_;
    size_t msgSize_;
    Sender* sender_;
    size_t periodicMsgCount_;
};

void runInSenderMode(Config & config, Topic const& topic, uint16_t msgSize, uint16_t startCore) {
    using namespace std;
    uint64_t cpuMask = (1ul << startCore);
    SingletonGuardian<HMBDC_LOGGER> logGuard(std::cout, cpuMask, "SCHED_OTHER");
#ifdef HMBDC_LOG_CONTEXT
    //only use a cpu when async logger is used
    cpuMask <<= 0x1u;
#endif
    SendCtx ctx; 
    SingletonGuardian<NetContext> g;

    auto engine = NetContext::instance().createSendTransportEngine(
        config, msgSize);
    SenderClient client(ctx, topic, msgSize);
    ctx.start(*engine, cpuMask, true);
    cpuMask <<= 0x1u;
    ctx.start(client, cpuMask);
    os::HandleSignals::onTermIntDo(
        [&ctx]() {
            ctx.stop();
        }
    );
    ctx.join();
}

using RecvCtx = Context<>;
struct ReceiverClient 
: Client<ReceiverClient, Message>
, TimerManager
, ReoccuringTimer {
    ReceiverClient(RecvCtx& ctx, size_t msgSize)
    : ReoccuringTimer(Duration::seconds(1u))
    , ctx_(ctx)
    , periodicMsgCount_(0)
    , seq_(0)
    , totalMsgReceived_(0)
    , missed_(0)
    , outOfOrder_(0)
    , duplicate_(0)
    , msgSize_(msgSize)
     {
        setCallback(
            [this](TimerManager& tm, SysTime const& now) {
            report();
            }
        );
        schedule(SysTime::now(), *this);
    }

    char const* hmbdcName() const { 
        return "perf-rx"; 
    }

    void handleMessageCb(Message const& r) {
        periodicMsgCount_++;
        if (seq_ != 0) {
            if (r.seq > seq_) {
            missed_ += r.seq - seq_ - 1;
            } else if (r.seq == seq_) {
                duplicate_++;
            } else {
                outOfOrder_++;
            }
        }
        seq_ = r.seq;
        totalMsgReceived_++;
    }

    void stoppedCb(std::exception const& e){
        std::cout << (e.what()) << std::endl;
    };

    void report() {
        HMBDC_LOG_n("msgSize=", msgSize_, " mps=", periodicMsgCount_
            , " missed=" , missed_ , '(' , missed_?missed_ * 100ul / (missed_ + totalMsgReceived_):0
            , "%) outOfOrder=" , outOfOrder_ 
            , " duplicate=" , duplicate_ 
            , " totalMsgCount=" , totalMsgReceived_ 
        );

        periodicMsgCount_ = 0u;
    }

    RecvCtx& ctx_;
    size_t periodicMsgCount_;
    size_t seq_;
    size_t totalMsgReceived_;
    size_t missed_;
    size_t outOfOrder_;
    size_t duplicate_;
    size_t msgSize_;
};

void runInReceiverMode(Config const& config, Topic const& topic, uint16_t msgSize, uint16_t inBuferSizePower2, uint16_t startCore) {   
    using namespace std;
    using namespace boost;
    SingletonGuardian<HMBDC_LOGGER> logGuard(std::cout);
    RecvCtx ctx(inBuferSizePower2, 1, msgSize);   

    SingletonGuardian<NetContext> g;
    auto& net = NetContext::instance();
    
    uint64_t cpuMask = (1ul << startCore);
    auto engine = net.createRecvTransportEngine(config, ctx.buffer());
    ReceiverClient client(ctx, msgSize);
    net.listenTo(topic);
    ctx.start(*engine, cpuMask, true);
    cpuMask <<= 0x1u;
    ctx.start(client, cpuMask, false);
    os::HandleSignals::onTermIntDo(
        [&ctx]() {
            ctx.stop();
        }
    );
    ctx.join();

    net.stopListenTo(topic);
}

int startNetPerf(char const* dftUserConfig, char const* helpStrIn, int argc, char** argv) {
    using namespace std;
    string role;
    string userCfg;
    uint16_t msgSize;
    uint16_t startCore;
    uint16_t inBuferSizePower2;
    Topic topic("t");
	if (argc > 1) role = string(argv[1]);
    if (role != string("--sender")
		&& role != string("--receiver")) {
			cerr << "1st arg needs to be either --sender or --receiver" << endl;
		exit(1);
	}
	argc--;
	argv++;

    bool isSender = role == string("--sender");
    namespace po = boost::program_options;
    po::options_description desc("additional allowed options (after --sender or --receiver)");
    string helpStr =
    "This program can be used to test messaging performance among a group of hosts.\n"
    "It can run as either a sender or a receiver. Normally one sender for a single interface in a test. \n"
    "When Sender sends out traffic, receivers subscribe to the traffic and collect stats.\n"
    "ctrl-c terminate.\n";
    helpStr += string(helpStrIn);
    desc.add_options()
    ("help", helpStr.c_str())
    ("cfg,c", po::value<std::string>(&userCfg), "use the user provided config file and ignore the ones in command options")
    ("print", "print effective cfg values")
    ("startCore,C", po::value<uint16_t>(&startCore)->default_value(0u), "the test uses 3 cores, specify starting core number")
    ("msgSize", po::value<uint16_t>(&msgSize)->default_value(8u), "size of the message (8-1024)")
    ("inBuferSizePower2", po::value<uint16_t>(&inBuferSizePower2)->default_value(0u), "2^inBufferSizePower2 is the number of message that can be buffered in the program, default 0 means automatically caulcated based on 32MB as lower bound")
    ("topic", po::value<Topic>(&topic)->default_value(topic), "topic used in the test");

    
    Config dft(dftUserConfig, isSender?"tx":"rx");
    auto params = dft.content();
    for (auto it = params.begin(); it != params.end();) {
    	string& name = it->first;
    	string& val = it->second;
    	it++;
    	string& comment = it->second;
    	it++;
    	desc.add_options()
    		(name.c_str(), po::value<string>(&val)->default_value(val), comment.c_str());
    }

    po::positional_options_description p;
    po::variables_map vm;
    po::store(po::command_line_parser(argc, argv).
        options(desc).positional(p).run(), vm);
    po::notify(vm);

    if (argc == 1 || vm.count("help") || msgSize < 8u) {
        cout << desc << "\n";
        return 0;
    }

    Config config;
    if (userCfg.size()) {
        std::istringstream cs(userCfg);
        config = Config(cs, isSender?"tx":"rx");
    } else {
        for (auto& p : params) {
        	config.put(p.first, p.second);
        }
    }

    if (vm.count("print")) {
        write_json(cout, config);
        exit(0);
    }

    if (isSender) {
        runInSenderMode(config, topic, msgSize, startCore);
    } else {
        if (!inBuferSizePower2) {
            inBuferSizePower2 = hmbdc::numeric::log2Upper(32ul * 1024ul * 1024ul / (8ul + msgSize));
            cout << "auto set --inBuferSizePower2=" << inBuferSizePower2 << endl;
        }
        runInReceiverMode(config, topic, msgSize, inBuferSizePower2, startCore);
    }
}

}}}

