#include "hmbdc/Copyright.hpp"
#pragma once

#include "hmbdc/os/Allocators.hpp"
#include <stdint.h>
#include <stddef.h>
#include <malloc.h>
#include <stdlib.h>
#include <memory.h>

#include <utility>
#include <stdexcept>

#include <cstddef>

#ifndef SMP_CACHE_BYTES
#define SMP_CACHE_BYTES 64
#endif

namespace hmbdc { namespace pattern {
namespace lf_misc {

struct DeadConsumer : std::runtime_error {
    DeadConsumer(const std::string& what_arg) 
    : std::runtime_error(what_arg){}
};

template <typename Seq>
struct chunk_base_ptr {
    template<typename Allocator = os::DefaultAllocator>
    chunk_base_ptr(uint32_t valueTypeSizePower2Num, uint32_t ringSizePower2Num
        , Allocator& allocator = os::DefaultAllocator::instance)
    : space_((char*)allocator.memalign(SMP_CACHE_BYTES, (1u << valueTypeSizePower2Num) * (1u << ringSizePower2Num))
            - (char*)this)
    , MASK((1u << ringSizePower2Num) - 1)
    , valueTypeSizePower2Num_(valueTypeSizePower2Num)
    , allocateFromHeap_(Allocator::fromHeap) {
    }

    static size_t footprint(uint32_t valueTypeSizePower2Num, uint32_t ringSizePower2Num) {
        return sizeof(chunk_base_ptr) + SMP_CACHE_BYTES * 2
            + (1u << valueTypeSizePower2Num) * (1u << ringSizePower2Num);
    }

    ~chunk_base_ptr() {
        if (allocateFromHeap_) {
            ::free(((char* const)this) + space_);
        }
    }

    void* operator + (size_t index) const __restrict__ {
        return (((char* const)this) + space_) + (index << valueTypeSizePower2Num_) + sizeof(Seq);
    }

    Seq* getSeq(size_t index) const __restrict__  {
        return reinterpret_cast<Seq*>((((char* const)this) + space_) + (index << valueTypeSizePower2Num_));
    }

    friend 
    size_t operator - (void const* __restrict__ from, chunk_base_ptr<Seq> const& __restrict__ start) {
        return (static_cast<char const*>(from) - sizeof(Seq) - (((char* const)&start) + start.space_)) 
            >> start.valueTypeSizePower2Num_;
    }
    std::ptrdiff_t const space_;
    Seq const MASK;
    uint32_t const valueTypeSizePower2Num_;
    bool allocateFromHeap_;
    chunk_base_ptr(chunk_base_ptr<Seq> const&) = delete;
    chunk_base_ptr& operator = (chunk_base_ptr<Seq> const&) = delete;
}  __attribute__((__aligned__(SMP_CACHE_BYTES)));

template <typename Seq>
struct iterator {
    using Sequence = Seq;
    chunk_base_ptr<Seq> const * __restrict__ start_;
    Seq seq_;

    //this ctor is only used by LockFreeBuffer itself
    iterator(chunk_base_ptr<Seq> const& __restrict__ start, Seq seq)
        : start_(&start), seq_(seq){}
    iterator() : start_(nullptr), seq_(0){}
    Seq seq() const {return seq_;}
    void clear() {start_ = nullptr;}

    iterator& operator ++() __restrict__ {
        ++seq_;
        return *this;
    }

    iterator operator ++(int) __restrict__ {
        iterator tmp = *this;
        ++*this;
        return tmp;
    }

    iterator operator + (size_t dis) const __restrict__ {
        iterator tmp = *this;
        tmp.seq_ += dis;
        return tmp;
    }

    size_t operator - (iterator const& other) const __restrict__ {
        return seq_ - other.seq_;
    }

    explicit operator bool() const __restrict__ {
        return start_;
    }

    bool operator < (iterator const& other) const __restrict__ {
        return seq_ < other.seq_;
    }

    bool operator == (iterator const& other) const __restrict__ {return seq_ == other.seq_;}
    bool operator != (iterator const& other) const __restrict__ {return seq_ != other.seq_;}
    void* operator*() const __restrict__ {return *start_ + (seq_ & start_->MASK);}
    template <typename T> T& get() const __restrict__ { return *static_cast<T*>(**this); }
    template <typename T>
    T* operator->() __restrict__ {return static_cast<T*>(*start_ + (seq_ & start_->MASK));}
};

} //lf_misc
}}